from cached_property import cached_property


class Parser:

    def __init__(self, path: str):
        self.path = path

    @cached_property
    def parsed(self):
        raise NotImplementedError

    @cached_property
    def contents(self):
        raise NotImplementedError

    @cached_property
    def docstring(self):
        raise NotImplementedError



from typing import Optional
import ast
import warnings
from cached_property import cached_property
from projectreport.analyzer.parsers.base import Parser


class PythonParser(Parser):

    @cached_property
    def parsed(self):
        try:
            return ast.parse(self.contents)
        except SyntaxError:
            warnings.warn(f'Could not parse {self.path} due to SyntaxError')
            return None

    @cached_property
    def contents(self) -> str:
        with open(self.path) as f:
            file_contents = f.read()
        return file_contents

    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parsed is None:
            return None
        return ast.get_docstring(self.parsed)



from typing import TYPE_CHECKING, Dict, Union, Sequence
from cached_property import cached_property
if TYPE_CHECKING:
    from projectreport.analyzer.project import Project
from projectreport.report.project import ProjectReport
from projectreport.report.base import BaseReport


class Report(BaseReport):
    """
    Pass Projects to create a report of the analysis of the projects.
    """

    def __init__(self, projects: Sequence['Project'], depth: int = 0):
        self.project_reports = [ProjectReport(project, depth=depth) for project in projects]

    @cached_property
    def data(self):
        return [report.data for report in self.project_reports]

    @cached_property
    def doc(self):
        from projectreport.report.latex import multi_project_latex_document
        return multi_project_latex_document(self.data)




from typing import List, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from projectreport.analyzer.project import Project
    from projectreport.analyzer.parsers.base import Parser
import os
from cached_property import cached_property

from projectreport.analyzer.analyzable import Analyzable
from projectreport.analyzer.analysis import ModuleAnalysis
from projectreport.analyzer.parsers.index import PARSER_EXTENSIONS


class Module(Analyzable):

    def __init__(self, path: str, package: Optional[str] = None, project: Optional['Project'] = None):
        base_path, extension = os.path.splitext(path)
        self.name = os.path.basename(base_path)
        self.extension = extension.strip('.')
        if package is None:
            package = self.name
        self.package = package
        super().__init__(path, project=project)
        self.analysis = ModuleAnalysis(self)

    @cached_property
    def parser(self) -> Optional['Parser']:
        if self.extension not in PARSER_EXTENSIONS:
            return None
        return PARSER_EXTENSIONS[self.extension](self.path)




from typing import Optional, Dict, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from projectreport.analyzer.project import Project
from cached_property import cached_property


class Analyzable:
    parser = None
    analysis = None

    def __init__(self, path: str, project: Optional['Project'] = None):
        self.path = path
        self.project = project

    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parser is None:
            return None
        return self.parser.docstring

    @cached_property
    def data(self) -> Dict[str, Union[str, int, dict]]:
        data = {}
        data.update(self.analysis.data)
        data.update(dict(
            docstring=self.docstring,
        ))
        if hasattr(self, 'name'):
            data.update(dict(name=self.name))
        return data







from typing import Optional, List, Sequence, TYPE_CHECKING
if TYPE_CHECKING:
    from projectreport.analyzer.module import Module
    from projectreport.analyzer.analyzable import Analyzable
    from projectreport.analyzer.folder import Folder
from copy import deepcopy
from datetime import datetime
import pygount
from cached_property import cached_property
import git


class Analysis:
    loc = None
    full_loc = None

    def __init__(self, analyzable: 'Analyzable'):
        self.git_analysis = GitAnalysis(analyzable)

    @cached_property
    def data(self):
        return dict(
            num_commits=self.git_analysis.num_commits,
            created=self.git_analysis.created,
            updated=self.git_analysis.updated,
            lines=self.loc,
            full_lines=self.full_loc,
            urls=self.git_analysis.urls
        )


class FolderAnalysis(Analysis):

    def __init__(self, folder: 'Folder'):
        self.lines = {key: 0 for key in [
            'code',
            'documentation',
            'empty',
            'string'
        ]}
        super().__init__(folder)

    def __repr__(self):
        return f'<PackageAnalysis(lines={self.lines})>'

    def add_module_analysis(self, analysis: 'ModuleAnalysis'):
        for attr in self.lines:
            value = getattr(analysis.source_analysis, attr)
            if value is None:
                value = 0
            self.lines[attr] += value

    def add_subfolder_analysis(self, analysis: 'FolderAnalysis'):
        for attr in self.lines:
            value = analysis.lines[attr]
            if value is None:
                value = 0
            self.lines[attr] += value

    @property
    def loc(self) -> int:
        return self.lines['code']

    @property
    def full_loc(self) -> int:
        loc = 0
        keys = [
            'code',
            'documentation',
            'empty',
            'string'
        ]
        for key in keys:
            loc += self.lines[key]
        return loc


class ModuleAnalysis(Analysis):

    def __init__(self, module: 'Module'):
        self.module = module
        self.source_analysis = pygount.source_analysis(self.module.path, self.module.package)
        self.loc = self.source_analysis.code

        full_loc = 0
        for key in ['code', 'documentation', 'empty', 'string']:
            full_loc += getattr(self.source_analysis, key, 0)
        self.full_loc = full_loc
        super().__init__(module)


class GitAnalysis:

    def __init__(self, analyzable: 'Analyzable'):
        self.ref = analyzable

    @cached_property
    def commits(self) -> Optional[List[git.Commit]]:
        if not self.has_repo:
            return None
        commits = [commit for commit in self.ref.project.repo.iter_commits(paths=self.ref.path)]
        return commits

    @cached_property
    def num_commits(self) -> Optional[int]:
        if not self.has_repo:
            return None
        return len(self.commits)

    @cached_property
    def created(self) -> Optional[datetime]:
        if not self.commits:
            return None
        return self.commits[-1].committed_datetime

    @cached_property
    def updated(self) -> Optional[datetime]:
        if not self.commits:
            return None
        return self.commits[0].committed_datetime

    @cached_property
    def has_repo(self) -> bool:
        return self.ref.project is not None and self.ref.project.repo is not None

    @cached_property
    def has_remote(self) -> bool:
        try:
            self.ref.project.repo.remote()
            return True
        except ValueError as e:
            if "Remote named 'origin' didn't exist" in str(e):
                # Repo does not have remote
                return False
            else:
                raise e

    @cached_property
    def urls(self) -> Optional[List[str]]:
        if not self.has_repo or not self.has_remote:
            return None

        urls = list(self.ref.project.repo.remote().urls)

        # Urls currently have .git on the end, strip
        urls = [url[:-4] for url in urls]
        return urls
